import torch
from torch import nn
from tqdm.auto import tqdm, trange
import matplotlib.pyplot as plt

from sentimentpl.models import SentimentPLModel
from sentimentpl.datautils import SentimentPLDataset


if __name__ == '__main__':
    batch_size = 32
    train_files = [#'sentiment_data/all.text.train.txt',
                   'data/all.sentence.train.txt']
    test_files = [#'sentiment_data/all.text.test.txt',
                  'data/all.sentence.test.txt']

    train_loader = torch.utils.data.DataLoader(SentimentPLDataset(train_files), batch_size, shuffle=True)
    test_loader = torch.utils.data.DataLoader(SentimentPLDataset(test_files), batch_size)

    model = SentimentPLModel().cuda()#tokenizer='trained_models/politicalBERT', embed_model='trained_models/politicalBERT').cuda()
    criterion = nn.MSELoss()
    optimizer = torch.optim.AdamW(model.parameters())
    n_epochs = 30
    train_embedding = [False]#*4 + [True]

    train_losses = []
    valid_losses = []

    best_val_loss = float('inf')

    for epoch in range(n_epochs):
        print('---------------------------------------------------------')
        tune_embedding = train_embedding[min(epoch, len(train_embedding)-1)]
        print(f'epoch {epoch+1}/{n_epochs}\n  Tune embedding: {tune_embedding}')
        # training
        total_loss = 0
        model.train()
        for i, (x, y) in tqdm(enumerate(train_loader), total=len(train_loader)):
            y = y.cuda()
            optimizer.zero_grad()
            pred = model(x, tune_embedding=tune_embedding)
            loss = criterion(pred, y)
            loss.backward()
            optimizer.step()
            total_loss += loss.item()
        train_losses.append(total_loss / (i + 1))

        # validation
        total_loss = 0
        model.eval()
        for i, (x, y) in enumerate(test_loader):
            y = y.cuda()
            with torch.no_grad():
                pred = model(x)
                loss = criterion(pred, y)
            total_loss += loss.item()
        valid_losses.append(total_loss / (i + 1))

        print(f'\tloss : {train_losses[-1]:.6f}, val_loss: {valid_losses[-1]:.6f}\n')

        # model saving
        if valid_losses[-1] <= best_val_loss:
            best_val_loss = valid_losses[-1]
            model.save()
            print('model saved!\n')

    print('Finished Training')

    plt.plot(train_losses)
    plt.plot(valid_losses)
    plt.legend(['train loss', 'valid loss'])
    plt.show()



from contextlib import ExitStack
import importlib
import torch
from torch import nn
from transformers import XLMTokenizer, RobertaModel

import io
try:
    import importlib.resources as pkg_resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as pkg_resources

from sentimentpl import trained_models

class SentimentPLModel(nn.Module):
    def __init__(self,
                 from_pretrained=None,
                 tokenizer="allegro/herbert-klej-cased-tokenizer-v1",
                 embed_model="allegro/herbert-klej-cased-v1"):
        super().__init__()

        self.tokenizer = XLMTokenizer.from_pretrained(tokenizer)
        self.embed_model = RobertaModel.from_pretrained(embed_model, return_dict=True)

        self.fc = nn.Sequential(nn.Dropout(0.5),
                                nn.Linear(768, 256), nn.ReLU(),
                                nn.Linear(256, 16), nn.ReLU(),
                                nn.Linear(16, 1), nn.Tanh())

        if from_pretrained is not None:
            f = io.BytesIO(importlib.resources.read_binary(trained_models, f'{from_pretrained}.pth'))
            self.fc.load_state_dict(torch.load(f))
            self.eval()

    def save(self, name='latest'):
        self.fc.to('cpu')
        torch.save(self.fc.state_dict(), f'./trained_models/{name}.pth')
        self.fc.to(next(self.embed_model.parameters()).device)

    def forward(self, x, tune_embedding=False):
        encoded = self.tokenizer(x, return_tensors='pt', padding=True)
        encoded = {k: v.to(next(self.parameters()).device) for k, v in encoded.items()}

        with ExitStack() as stack:
            if not tune_embedding:
                stack.enter_context(torch.no_grad())
            embeddings = self.embed_model(**encoded)['pooler_output'].float()

        return self.fc(embeddings)



from transformers import XLMTokenizer, RobertaModel
from tqdm.auto import tqdm, trange
import torch



mapping = {'plus': 1., 'minus': -1., 'zero': 0., 'amb': 0.}

def load_embed(filenames):
    lines = []
    for filename in filenames:
        with open(filename, 'r', encoding='utf-8') as f:
            lines += [line for line in f]
    X, Y = [], []
    for line in lines:
        if '__label__' in line:
            text, sentiment = line.split('__label__')
            labels = [torch.tensor(v).unsqueeze(-1).float() for k, v in mapping.items() if k in sentiment][0]
            X.append(text)
            Y.append(labels)
    return X, Y


class SentimentPLDataset(torch.utils.data.Dataset):
    def __init__(self, filename):
        self.X, self.y = load_embed(filename)

    def __len__(self):
        return len(self.X)

    def __getitem__(self, i):
        return self.X[i], self.y[i]



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="sentimentpl",
    version="0.0.6",
    author="Filip Strzałka",
    author_email="strzalkafilip@gmail.com",
    description="PyTorch models for polish language sentiment regression based on allegro/herbert and CLARIN-PL dataset",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/philvec/sentimentPL",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=['torch>=1.7.1', 'transformers', 'tqdm', 'matplotlib']
)
